// LISTING 4 --- CSHELP.CPP
// Context-sensitive help system.  All this code is hidden away.
// The only thing the user sees is CSHELP.HPP. Notice the use of
// "static" to prevent "global namespace pollution."
// by Bruce Eckel, Revolution2
// (c) 1990 C Gazette. Use freely if authorship is acknowledged

#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include "lifo.hpp"
#include "cshelp.hpp"

/////////////////////  Direct Video Control ////////////////////////
// Function to get the video mode:
static int getmode() { // From "DOS Programmer's Reference, 2nd ed" Pg 329
  REGS regs;
  regs.h.ah = 0x0f;
  int86(0x10, &regs, &regs); // Video Interrupt
  return regs.h.al;
}

// Establish video base address as mono or color:
static int BASE = ((getmode() == 7) ? 0xb000 : 0xb800);

// direct screen write:
inline void putc_at(char ch, int x, int y) {
  pokeb(BASE,((y * 80) + x) * 2,ch);
}

static void put_screen(char * screen[]) {
  int i = 0;
  int row= 0, column = 0;
  while(*screen[i]) {        // empty line means the end
    char * p = screen[i++];  // point to beginning of current line
    while(*p)
      putc_at(*p++, column++, row);
    row++;
    column = 0;
  }
}

// Figure out size of screen & create storage space for it.  Note the
// use of the comma operator for sequential evaluation:
static text_info ti;  // Borland structure in conio.h
static char * old_screen =
  new char[(gettextinfo(&ti), 2 * ti.screenheight * ti.screenwidth)];

//////////////////////// Interrupt Control /////////////////////////
// Hardware constants, collected together into an untagged enum:
static enum {
  kb_int =  0x09 , // keyboard interrupt
  dos_bg = 0x28,   // DOS "background process"
  kb_data = 0x60,  // PIO port A -- contains keyboard scan code
  IC8259 = 0x20,   // 8259 i/o location
  EOI = 0x20,      // 8259 end-of-interrupt command
  hotkey = 0x3b,   // f1 key
  quitkey = 0x3c   // f2 key
};
// Semaphores for communication between processes:
static int run_help = 0;
static int help_running = 0;
static int charflag = 0;
static unsigned char bg_char = 0;

// Pointers to functions:
static void interrupt (*oldkb_handler)(...);
static void interrupt (*oldbg_process)(...);
// declare the interrupt functions:
static void interrupt kb_handler(...);
static void interrupt bg_process(...);
// variables to hold the old cursor screen position:
static int oldx;
static int oldy;

//////////////////////// The Help Stack ////////////////////////////
struct helpnode : node { // inherit from lifo node
  char ** helptext;  // add data to the node
  helpnode(char ** helptxt, node *n) : helptext(helptxt), node(n) {}
};

// This screen is always installed when the help_stack is created, to
// insure SOMETHING happens when the hot key is pressed.
static char * default_help_screen[] = {
"default help screen",
"F2 to quit",
"" // terminator
};

class help_stack : public lifo {
  helpnode * current;  // currently being displayed
  int default_flag; // if 1, default_help_screen is being used
public:
  helpnode * top() { return current = (helpnode *)lifo::top(); }
  helpnode * next() { return (helpnode *)lifo::next(); }
  // New type of "push":
  void push(char ** hs) {
    if(default_flag) {
      lifo::pop();  // remove default_help_screen
      default_flag = 0;
    }
    head = new helpnode(hs, head);
  }
  int pop() {
    if(!lifo::pop()) {           // if we just emptied the stack...
      push(default_help_screen); // put the default screen back on
      default_flag = 1;
    }
    return 1;                    // This stack never empties.
  }

  help_stack() {                 // constructor
    push(default_help_screen);
    default_flag = 1;
    current = top();
    // save previous interrupt handlers & install new ones:
    oldkb_handler = getvect(kb_int);
    setvect(kb_int, kb_handler);
    oldbg_process = getvect(dos_bg);
    setvect(dos_bg, bg_process);
  }
  ~help_stack() {  // destructor
    // restore previous interrupt handlers:
    setvect(dos_bg, oldbg_process);
    setvect(kb_int, oldkb_handler);
  }
  void show_help_screen() {
    put_screen(current->helptext);
  }
  void up_screen() {
    helpnode * tmp = next();
    current = (tmp ? tmp : current);  // move down on help stack if not NULL
    show_help_screen();
  }
};

static help_stack __hs__;  // single static instance

// Help handler definitions:
help_handler::help_handler(char ** help_screen) {
  __hs__.push(help_screen);
}

help_handler::~help_handler() { __hs__.pop(); }

///////////////////  The Interrupt Handlers ////////////////////////
static void interrupt kb_handler(...) {
  if(help_running) {            // capture keystrokes for bg_process
    bg_char = inportb(kb_data); // read keyboard data
    charflag++;  // set semaphore so bg_process knows a key was pressed
    outportb(IC8259,EOI);  // Tell the 8259 to quit sending the interrupt
    // (8259 control from "IBM PC From the Inside Out")
    return;
  }
  if(inportb(kb_data) == hotkey) {
    run_help++;  // set semaphore to tell bg_process to run
    outportb(IC8259,EOI);  // Tell the 8259 to quit sending the interrupt
    return;
  }
  oldkb_handler(); // no other conditions -- normal keystroke
}

// DOS "background process":
static void interrupt bg_process(...) {
  if (run_help) {      // semaphore says help was requested
    help_running = 1;  // hold off further setting of run_help
    run_help = 0;      // clear semaphore
    // save the old screen:
    gettext(1,1, ti.screenwidth, ti.screenheight, old_screen);
    oldx = wherex();  // and the old cursor coordinates
    oldy = wherey();
    clrscr();
    __hs__.top();     // display the current help screen
    __hs__.show_help_screen();
  }
  if (charflag) {    // key pressed inside of help
    charflag = 0;    // clear semaphore.  Now test for keys:
    if(bg_char == quitkey) { // quit from help
      // restore old screen:
      puttext(1,1, ti.screenwidth, ti.screenheight, old_screen);
      gotoxy(oldx, oldy);  // and old cursor position
      help_running = 0; // clear semaphore to allow setting of run_help
    }
    if(bg_char == hotkey) { // go up one help level
      clrscr();
      __hs__.up_screen();
    }
  }
  oldbg_process();  // always call other background processes
}